@@ -27,95 +27,19 @@
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
+#include <assert.h>
 
 #include "pixman-private.h"
 
 #define Alpha(x) ((x) >> 24)
 
-static source_pict_class_t
-SourcePictureClassify (pixman_image_t *image,
-		       int	       x,
-		       int	       y,
-		       int	       width,
-		       int	       height)
-{
-    source_image_t *pict = &image->source;
-    
-    pict->class = SOURCE_IMAGE_CLASS_UNKNOWN;
-    
-    if (pict->common.type == SOLID)
-    {
-	pict->class = SOURCE_IMAGE_CLASS_HORIZONTAL;
-    }
-    else if (pict->common.type == LINEAR)
-    {
-	linear_gradient_t *linear = (linear_gradient_t *)pict;
-	pixman_vector_t   v;
-	pixman_fixed_32_32_t l;
-	pixman_fixed_48_16_t dx, dy, a, b, off;
-	pixman_fixed_48_16_t factors[4];
-	int	     i;
-
-	dx = linear->p2.x - linear->p1.x;
-	dy = linear->p2.y - linear->p1.y;
-	l = dx * dx + dy * dy;
-	if (l)
-	{
-	    a = (dx << 32) / l;
-	    b = (dy << 32) / l;
-	}
-	else
-	{
-	    a = b = 0;
-	}
-
-	off = (-a * linear->p1.x
-	       -b * linear->p1.y) >> 16;
-
-	for (i = 0; i < 3; i++)
-	{
-	    v.vector[0] = pixman_int_to_fixed ((i % 2) * (width  - 1) + x);
-	    v.vector[1] = pixman_int_to_fixed ((i / 2) * (height - 1) + y);
-	    v.vector[2] = pixman_fixed_1;
-
-	    if (pict->common.transform)
-	    {
-		if (!pixman_transform_point_3d (pict->common.transform, &v))
-		{
-		    pict->class = SOURCE_IMAGE_CLASS_UNKNOWN;
-		    goto out;
-		}
-	    }
-
-	    factors[i] = ((a * v.vector[0] + b * v.vector[1]) >> 16) + off;
-	}
-
-	if (factors[2] == factors[0])
-	    pict->class = SOURCE_IMAGE_CLASS_HORIZONTAL;
-	else if (factors[1] == factors[0])
-	    pict->class = SOURCE_IMAGE_CLASS_VERTICAL;
-    }
-
-out:
-    return pict->class;
-}
-
-static void
-init_source_image (source_image_t *image)
-{
-    image->class = SOURCE_IMAGE_CLASS_UNKNOWN;
-    image->common.classify = SourcePictureClassify;
-}
-
-static pixman_bool_t
-init_gradient (gradient_t     *gradient,
-	       const pixman_gradient_stop_t *stops,
-	       int	       n_stops)
+pixman_bool_t
+_pixman_init_gradient (gradient_t     *gradient,
+		       const pixman_gradient_stop_t *stops,
+		       int	       n_stops)
 {
     return_val_if_fail (n_stops > 0, FALSE);
 
-    init_source_image (&gradient->common);
-
     gradient->stops = pixman_malloc_ab (n_stops, sizeof (pixman_gradient_stop_t));
     if (!gradient->stops)
 	return FALSE;
@@ -127,22 +51,45 @@ init_gradient (gradient_t     *gradient,
     gradient->stop_range = 0xffff;
     gradient->color_table = NULL;
     gradient->color_table_size = 0;
+    gradient->common.class = SOURCE_IMAGE_CLASS_UNKNOWN;
 
     return TRUE;
 }
 
-static uint32_t
-color_to_uint32 (const pixman_color_t *color)
+/*
+ * By default, just evaluate the image at 32bpp and expand.  Individual image
+ * types can plug in a better scanline getter if they want to. For example
+ * we  could produce smoother gradients by evaluating them at higher color depth, but
+ * that's a project for the future.
+ */
+void
+_pixman_image_get_scanline_64_generic (pixman_image_t * pict, int x, int y, int width,
+				       uint64_t *buffer, uint64_t *mask, uint32_t maskBits)
 {
-    return
-	(color->alpha >> 8 << 24) |
-	(color->red >> 8 << 16) |
-        (color->green & 0xff00) |
-	(color->blue >> 8);
+    uint32_t *mask8 = NULL;
+
+    // Contract the mask image, if one exists, so that the 32-bit fetch function
+    // can use it.
+    if (mask) {
+        mask8 = pixman_malloc_ab(width, sizeof(uint32_t));
+	if (!mask8)
+	    return;
+	
+        pixman_contract(mask8, mask, width);
+    }
+
+    // Fetch the source image into the first half of buffer.
+    _pixman_image_get_scanline_32 (pict, x, y, width, (uint32_t*)buffer, mask8,
+				   maskBits);
+
+    // Expand from 32bpp to 64bpp in place.
+    pixman_expand(buffer, (uint32_t*)buffer, PIXMAN_a8r8g8b8, width);
+
+    free(mask8);
 }
 
-static pixman_image_t *
-allocate_image (void)
+pixman_image_t *
+_pixman_image_allocate (void)
 {
     pixman_image_t *image = malloc (sizeof (pixman_image_t));
 
@@ -150,10 +97,10 @@ allocate_image (void)
     {
 	image_common_t *common = &image->common;
 
-	pixman_region32_init (&common->full_region);
 	pixman_region32_init (&common->clip_region);
-	common->src_clip = &common->full_region;
-	common->has_client_clip = FALSE;
+
+	common->have_clip_region = FALSE;
+	common->clip_sources = FALSE;
 	common->transform = NULL;
 	common->repeat = PIXMAN_REPEAT_NONE;
 	common->filter = PIXMAN_FILTER_NEAREST;
@@ -165,6 +112,9 @@ allocate_image (void)
 	common->read_func = NULL;
 	common->write_func = NULL;
 	common->classify = NULL;
+	common->client_clip = FALSE;
+	common->destroy_func = NULL;
+	common->destroy_data = NULL;
     }
 
     return image;
@@ -183,147 +133,45 @@ _pixman_image_classify (pixman_image_t *image,
 	return SOURCE_IMAGE_CLASS_UNKNOWN;
 }
 
-#define READ_ACCESS(f) ((image->common.read_func)? f##_accessors : f)
-
-static void fbFetchSolid(bits_image_t * image, int x, int y, int width, uint32_t *buffer, uint32_t *mask, uint32_t maskBits)
+void
+_pixman_image_get_scanline_32 (pixman_image_t *image, int x, int y, int width,
+			       uint32_t *buffer, uint32_t *mask, uint32_t mask_bits)
 {
-    uint32_t color;
-    uint32_t *end;
-    fetchPixelProc32 fetch = READ_ACCESS(pixman_fetchPixelProcForPicture32)(image);
-
-    color = fetch(image, 0, 0);
-
-    end = buffer + width;
-    while (buffer < end)
-	*(buffer++) = color;
+    image->common.get_scanline_32 (image, x, y, width, buffer, mask, mask_bits);
 }
 
-static void fbFetchSolid64(bits_image_t * image, int x, int y, int width, uint64_t *buffer, void *unused, uint32_t unused2)
+void
+_pixman_image_get_scanline_64 (pixman_image_t *image, int x, int y, int width,
+			       uint32_t *buffer, uint32_t *unused, uint32_t unused2)
 {
-    uint64_t color;
-    uint64_t *end;
-    fetchPixelProc64 fetch = READ_ACCESS(pixman_fetchPixelProcForPicture64)(image);
-
-    color = fetch(image, 0, 0);
-
-    end = buffer + width;
-    while (buffer < end)
-	*(buffer++) = color;
+    image->common.get_scanline_64 (image, x, y, width, buffer, unused, unused2);
 }
 
-static void fbFetch(bits_image_t * image, int x, int y, int width, uint32_t *buffer, uint32_t *mask, uint32_t maskBits)
-{
-    fetchProc32 fetch = READ_ACCESS(pixman_fetchProcForPicture32)(image);
-
-    fetch(image, x, y, width, buffer);
-}
-
-static void fbFetch64(bits_image_t * image, int x, int y, int width, uint64_t *buffer, void *unused, uint32_t unused2)
-{
-    fetchProc64 fetch = READ_ACCESS(pixman_fetchProcForPicture64)(image);
-
-    fetch(image, x, y, width, buffer);
-}
+/* Even thought the type of buffer is uint32_t *, the function actually expects
+ * a uint64_t *buffer.
+ */
 
 scanFetchProc
 _pixman_image_get_fetcher (pixman_image_t *image,
 			   int             wide)
 {
-    if (IS_SOURCE_IMAGE (image))
-    {
-	if (wide)
-	    return (scanFetchProc)pixmanFetchSourcePict64;
-	else
-	    return (scanFetchProc)pixmanFetchSourcePict;
-    }
+    assert (image->common.get_scanline_64);
+    assert (image->common.get_scanline_32);
+    
+    if (wide)
+	return image->common.get_scanline_64;
     else
-    {
-	bits_image_t *bits = (bits_image_t *)image;
-
-	if (bits->common.alpha_map)
-	{
-	    if (wide)
-		return (scanFetchProc)READ_ACCESS(fbFetchExternalAlpha64);
-	    else
-		return (scanFetchProc)READ_ACCESS(fbFetchExternalAlpha);
-	}
-	else if ((bits->common.repeat != PIXMAN_REPEAT_NONE) &&
-		 bits->width == 1 &&
-		 bits->height == 1)
-	{
-	    if (wide)
-		return (scanFetchProc)fbFetchSolid64;
-	    else
-		return (scanFetchProc)fbFetchSolid;
-	}
-	else if (!bits->common.transform && bits->common.filter != PIXMAN_FILTER_CONVOLUTION
-                && bits->common.repeat != PIXMAN_REPEAT_PAD && bits->common.repeat != PIXMAN_REPEAT_REFLECT)
-	{
-	    if (wide)
-		return (scanFetchProc)fbFetch64;
-	    else
-		return (scanFetchProc)fbFetch;
-	}
-	else
-	{
-	    if (wide)
-		return (scanFetchProc)READ_ACCESS(fbFetchTransformed64);
-	    else
-		return (scanFetchProc)READ_ACCESS(fbFetchTransformed);
-	}
-    }
+	return image->common.get_scanline_32;
 }
 
-
-
 #define WRITE_ACCESS(f) ((image->common.write_func)? f##_accessors : f)
 
 static void
-fbStore(bits_image_t * image, int x, int y, int width, uint32_t *buffer)
+image_property_changed (pixman_image_t *image)
 {
-    uint32_t *bits;
-    int32_t stride;
-    storeProc32 store = WRITE_ACCESS(pixman_storeProcForPicture32)(image);
-    const pixman_indexed_t * indexed = image->indexed;
-
-    bits = image->bits;
-    stride = image->rowstride;
-    bits += y*stride;
-    store((pixman_image_t *)image, bits, buffer, x, width, indexed);
-}
-
-static void
-fbStore64(bits_image_t * image, int x, int y, int width, uint64_t *buffer)
-{
-    uint32_t *bits;
-    int32_t stride;
-    storeProc64 store = WRITE_ACCESS(pixman_storeProcForPicture64)(image);
-    const pixman_indexed_t * indexed = image->indexed;
-
-    bits = image->bits;
-    stride = image->rowstride;
-    bits += y*stride;
-    store((pixman_image_t *)image, bits, buffer, x, width, indexed);
-}
-
-scanStoreProc
-_pixman_image_get_storer (pixman_image_t *image,
-			  int             wide)
-{
-    if (image->common.alpha_map)
-    {
-	if (wide)
-	    return (scanStoreProc)WRITE_ACCESS(fbStoreExternalAlpha64);
-	else
-	    return (scanStoreProc)WRITE_ACCESS(fbStoreExternalAlpha);
-    }
-    else
-    {
-	if (wide)
-	    return (scanStoreProc)fbStore64;
-	else
-	    return (scanStoreProc)fbStore;
-    }
+    
+    
+    image->common.property_changed (image);
 }
 
 /* Ref Counting */
@@ -345,8 +193,10 @@ pixman_image_unref (pixman_image_t *image)
 
     if (common->ref_count == 0)
     {
+	if (image->common.destroy_func)
+	    image->common.destroy_func (image, image->common.destroy_data);
+	
 	pixman_region32_fini (&common->clip_region);
-	pixman_region32_fini (&common->full_region);
 
 	if (common->transform)
 	    free (common->transform);
@@ -383,233 +233,22 @@ pixman_image_unref (pixman_image_t *image)
     return FALSE;
 }
 
+PIXMAN_EXPORT void
+pixman_image_set_destroy_function (pixman_image_t *image,
+				   pixman_image_destroy_func_t func,
+				   void *data)
+{
+    image->common.destroy_func = func;
+    image->common.destroy_data = data;
+}
+			       
+
 /* Constructors */
-PIXMAN_EXPORT pixman_image_t *
-pixman_image_create_solid_fill (pixman_color_t *color)
+
+void
+_pixman_image_reset_clip_region (pixman_image_t *image)
 {
-    pixman_image_t *img = allocate_image();
-    if (!img)
-	return NULL;
-
-    init_source_image (&img->solid.common);
-
-    img->type = SOLID;
-    img->solid.color = color_to_uint32 (color);
-
-    return img;
-}
-
-PIXMAN_EXPORT pixman_image_t *
-pixman_image_create_linear_gradient (pixman_point_fixed_t         *p1,
-				     pixman_point_fixed_t         *p2,
-				     const pixman_gradient_stop_t *stops,
-				     int                           n_stops)
-{
-    pixman_image_t *image;
-    linear_gradient_t *linear;
-
-    return_val_if_fail (n_stops >= 2, NULL);
-
-    image = allocate_image();
-
-    if (!image)
-	return NULL;
-
-    linear = &image->linear;
-
-    if (!init_gradient (&linear->common, stops, n_stops))
-    {
-	free (image);
-	return NULL;
-    }
-
-    linear->p1 = *p1;
-    linear->p2 = *p2;
-
-    image->type = LINEAR;
-
-    return image;
-}
-
-
-PIXMAN_EXPORT pixman_image_t *
-pixman_image_create_radial_gradient (pixman_point_fixed_t         *inner,
-				     pixman_point_fixed_t         *outer,
-				     pixman_fixed_t                inner_radius,
-				     pixman_fixed_t                outer_radius,
-				     const pixman_gradient_stop_t *stops,
-				     int                           n_stops)
-{
-    pixman_image_t *image;
-    radial_gradient_t *radial;
-
-    return_val_if_fail (n_stops >= 2, NULL);
-
-    image = allocate_image();
-
-    if (!image)
-	return NULL;
-
-    radial = &image->radial;
-
-    if (!init_gradient (&radial->common, stops, n_stops))
-    {
-	free (image);
-	return NULL;
-    }
-
-    image->type = RADIAL;
-
-    radial->c1.x = inner->x;
-    radial->c1.y = inner->y;
-    radial->c1.radius = inner_radius;
-    radial->c2.x = outer->x;
-    radial->c2.y = outer->y;
-    radial->c2.radius = outer_radius;
-    radial->cdx = pixman_fixed_to_double (radial->c2.x - radial->c1.x);
-    radial->cdy = pixman_fixed_to_double (radial->c2.y - radial->c1.y);
-    radial->dr = pixman_fixed_to_double (radial->c2.radius - radial->c1.radius);
-    radial->A = (radial->cdx * radial->cdx
-		 + radial->cdy * radial->cdy
-		 - radial->dr  * radial->dr);
-
-    return image;
-}
-
-PIXMAN_EXPORT pixman_image_t *
-pixman_image_create_conical_gradient (pixman_point_fixed_t *center,
-				      pixman_fixed_t angle,
-				      const pixman_gradient_stop_t *stops,
-				      int n_stops)
-{
-    pixman_image_t *image = allocate_image();
-    conical_gradient_t *conical;
-
-    if (!image)
-	return NULL;
-
-    conical = &image->conical;
-
-    if (!init_gradient (&conical->common, stops, n_stops))
-    {
-	free (image);
-	return NULL;
-    }
-
-    image->type = CONICAL;
-    conical->center = *center;
-    conical->angle = angle;
-
-    return image;
-}
-
-static uint32_t *
-create_bits (pixman_format_code_t format,
-	     int		  width,
-	     int		  height,
-	     int		 *rowstride_bytes)
-{
-    int stride;
-    int buf_size;
-    int bpp;
-
-    /* what follows is a long-winded way, avoiding any possibility of integer
-     * overflows, of saying:
-     * stride = ((width * bpp + FB_MASK) >> FB_SHIFT) * sizeof (uint32_t);
-     */
-
-    bpp = PIXMAN_FORMAT_BPP (format);
-    if (pixman_multiply_overflows_int (width, bpp))
-	return NULL;
-
-    stride = width * bpp;
-    if (pixman_addition_overflows_int (stride, FB_MASK))
-	return NULL;
-
-    stride += FB_MASK;
-    stride >>= FB_SHIFT;
-
-#if FB_SHIFT < 2
-    if (pixman_multiply_overflows_int (stride, sizeof (uint32_t)))
-	return NULL;
-#endif
-    stride *= sizeof (uint32_t);
-
-    if (pixman_multiply_overflows_int (height, stride))
-	return NULL;
-
-    buf_size = height * stride;
-
-    if (rowstride_bytes)
-	*rowstride_bytes = stride;
-
-    return calloc (buf_size, 1);
-}
-
-static void
-reset_clip_region (pixman_image_t *image)
-{
-    pixman_region32_fini (&image->common.clip_region);
-
-    if (image->type == BITS)
-    {
-	pixman_region32_init_rect (&image->common.clip_region, 0, 0,
-				   image->bits.width, image->bits.height);
-    }
-    else
-    {
-	pixman_region32_init (&image->common.clip_region);
-    }
-}
-
-PIXMAN_EXPORT pixman_image_t *
-pixman_image_create_bits (pixman_format_code_t  format,
-			  int                   width,
-			  int                   height,
-			  uint32_t	       *bits,
-			  int			rowstride_bytes)
-{
-    pixman_image_t *image;
-    uint32_t *free_me = NULL;
-
-    /* must be a whole number of uint32_t's
-     */
-    return_val_if_fail (bits == NULL ||
-			(rowstride_bytes % sizeof (uint32_t)) == 0, NULL);
-
-    if (!bits && width && height)
-    {
-	free_me = bits = create_bits (format, width, height, &rowstride_bytes);
-	if (!bits)
-	    return NULL;
-    }
-
-    image = allocate_image();
-
-    if (!image) {
-	if (free_me)
-	    free (free_me);
-	return NULL;
-    }
-    
-    image->type = BITS;
-    image->bits.format = format;
-    image->bits.width = width;
-    image->bits.height = height;
-    image->bits.bits = bits;
-    image->bits.free_me = free_me;
-
-    image->bits.rowstride = rowstride_bytes / (int) sizeof (uint32_t); /* we store it in number
-								  * of uint32_t's
-								  */
-    image->bits.indexed = NULL;
-
-    pixman_region32_fini (&image->common.full_region);
-    pixman_region32_init_rect (&image->common.full_region, 0, 0,
-			       image->bits.width, image->bits.height);
-
-    reset_clip_region (image);
-    return image;
+    image->common.have_clip_region = FALSE;
 }
 
 PIXMAN_EXPORT pixman_bool_t
@@ -617,17 +256,23 @@ pixman_image_set_clip_region32 (pixman_image_t *image,
 				pixman_region32_t *region)
 {
     image_common_t *common = (image_common_t *)image;
+    pixman_bool_t result;
 
     if (region)
     {
-	return pixman_region32_copy (&common->clip_region, region);
+	if ((result = pixman_region32_copy (&common->clip_region, region)))
+	    image->common.have_clip_region = TRUE;
     }
     else
     {
-	reset_clip_region (image);
+	_pixman_image_reset_clip_region (image);
 
-	return TRUE;
+	result = TRUE;
     }
+
+    image_property_changed (image);
+
+    return result;
 }
 
 
@@ -636,26 +281,30 @@ pixman_image_set_clip_region (pixman_image_t    *image,
 			      pixman_region16_t *region)
 {
     image_common_t *common = (image_common_t *)image;
+    pixman_bool_t result;
 
     if (region)
     {
-	return pixman_region32_copy_from_region16 (&common->clip_region, region);
+	if ((result = pixman_region32_copy_from_region16 (&common->clip_region, region)))
+	    image->common.have_clip_region = TRUE;
     }
     else
     {
-	reset_clip_region (image);
+	_pixman_image_reset_clip_region (image);
 
-	return TRUE;
+	result = TRUE;
     }
+
+    image_property_changed (image);
+
+    return result;
 }
 
-/* Sets whether the clip region includes a clip region set by the client
- */
 PIXMAN_EXPORT void
 pixman_image_set_has_client_clip (pixman_image_t *image,
 				  pixman_bool_t	  client_clip)
 {
-    image->common.has_client_clip = client_clip;
+    image->common.client_clip = client_clip;
 }
 
 PIXMAN_EXPORT pixman_bool_t
@@ -671,6 +320,7 @@ pixman_image_set_transform (pixman_image_t           *image,
     };
 
     image_common_t *common = (image_common_t *)image;
+    pixman_bool_t result;
 
     if (common->transform == transform)
 	return TRUE;
@@ -679,16 +329,24 @@ pixman_image_set_transform (pixman_image_t           *image,
     {
 	free(common->transform);
 	common->transform = NULL;
-	return TRUE;
+	result = TRUE;
+	goto out;
     }
 
     if (common->transform == NULL)
 	common->transform = malloc (sizeof (pixman_transform_t));
+
     if (common->transform == NULL)
-	return FALSE;
+    {
+	result = FALSE;
+	goto out;
+    }
 
     memcpy(common->transform, transform, sizeof(pixman_transform_t));
 
+out:
+    image_property_changed (image);
+    
     return TRUE;
 }
 
@@ -697,6 +355,8 @@ pixman_image_set_repeat (pixman_image_t  *image,
 			 pixman_repeat_t  repeat)
 {
     image->common.repeat = repeat;
+
+    image_property_changed (image);
 }
 
 PIXMAN_EXPORT pixman_bool_t
@@ -729,19 +389,16 @@ pixman_image_set_filter (pixman_image_t       *image,
 
     common->filter_params = new_params;
     common->n_filter_params = n_params;
+
+    image_property_changed (image);
     return TRUE;
 }
 
 PIXMAN_EXPORT void
 pixman_image_set_source_clipping (pixman_image_t  *image,
-				  pixman_bool_t    source_clipping)
+				  pixman_bool_t    clip_sources)
 {
-    image_common_t *common = &image->common;
-
-    if (source_clipping)
-	common->src_clip = &common->clip_region;
-    else
-	common->src_clip = &common->full_region;
+    image->common.clip_sources = clip_sources;
 }
 
 /* Unlike all the other property setters, this function does not
@@ -755,6 +412,8 @@ pixman_image_set_indexed (pixman_image_t	 *image,
     bits_image_t *bits = (bits_image_t *)image;
 
     bits->indexed = indexed;
+
+    image_property_changed (image);
 }
 
 PIXMAN_EXPORT void
@@ -780,6 +439,8 @@ pixman_image_set_alpha_map (pixman_image_t *image,
 
     common->alpha_origin.x = x;
     common->alpha_origin.y = y;
+
+    image_property_changed (image);
 }
 
 PIXMAN_EXPORT void
@@ -787,6 +448,8 @@ pixman_image_set_component_alpha   (pixman_image_t       *image,
 				    pixman_bool_t         component_alpha)
 {
     image->common.component_alpha = component_alpha;
+
+    image_property_changed (image);
 }
 
 
@@ -799,6 +462,8 @@ pixman_image_set_accessors (pixman_image_t             *image,
 
     image->common.read_func = read_func;
     image->common.write_func = write_func;
+
+    image_property_changed (image);
 }
 
 PIXMAN_EXPORT uint32_t *
@@ -846,6 +511,16 @@ pixman_image_get_depth (pixman_image_t *image)
     return 0;
 }
 
+static uint32_t
+color_to_uint32 (const pixman_color_t *color)
+{
+    return
+	(color->alpha >> 8 << 24) |
+	(color->red >> 8 << 16) |
+        (color->green & 0xff00) |
+	(color->blue >> 8);
+}
+
 static pixman_bool_t
 color_to_pixel (pixman_color_t *color,
 		uint32_t       *pixel,
@@ -938,11 +613,14 @@ pixman_image_fill_rectangles (pixman_op_t		    op,
 		pixman_box32_t *boxes;
 
 		pixman_region32_init_rect (&fill_region, rects[i].x, rects[i].y, rects[i].width, rects[i].height);
-		if (!pixman_region32_intersect (&fill_region,
-						&fill_region,
-						&dest->common.clip_region))
-		    return FALSE;
 
+		if (dest->common.have_clip_region)
+		{
+		    if (!pixman_region32_intersect (&fill_region,
+						    &fill_region,
+						    &dest->common.clip_region))
+			return FALSE;
+		}
 
 		boxes = pixman_region32_rectangles (&fill_region, &n_boxes);
 		for (j = 0; j < n_boxes; ++j)
@@ -1013,57 +691,51 @@ pixman_image_can_get_solid (pixman_image_t *image)
 }
 
 pixman_bool_t
-pixman_image_is_opaque(pixman_image_t *image)
+pixman_image_is_opaque (pixman_image_t *image)
 {
-    int i = 0;
-    int gradientNumberOfColors = 0;
+    int i;
 
-    if(image->common.alpha_map)
+    if (image->common.alpha_map)
         return FALSE;
 
-    switch(image->type)
+    switch (image->type)
     {
     case BITS:
-        if(PIXMAN_FORMAT_A(image->bits.format))
+	if (image->common.repeat == PIXMAN_REPEAT_NONE)
+	    return FALSE;
+	
+        if (PIXMAN_FORMAT_A (image->bits.format))
             return FALSE;
         break;
 
     case LINEAR:
-    case CONICAL:
     case RADIAL:
-        gradientNumberOfColors = image->gradient.n_stops;
-        i=0;
-        while(i<gradientNumberOfColors)
-        {
-            if(image->gradient.stops[i].color.alpha != 0xffff)
+	if (image->common.repeat == PIXMAN_REPEAT_NONE)
+	    return FALSE;
+	
+	for (i = 0; i < image->gradient.n_stops; ++i)
+	{
+            if (image->gradient.stops[i].color.alpha != 0xffff)
                 return FALSE;
-            i++;
         }
         break;
 
+    case CONICAL:
+	/* Conical gradients always have a transparent border */
+	return FALSE;
+	break;
+	
     case SOLID:
-         if(Alpha(image->solid.color) != 0xff)
+         if (Alpha (image->solid.color) != 0xff)
             return FALSE;
         break;
     }
 
-    /* Convolution filters can introduce translucency if the sum of the weights
-       is lower than 1. */
+    /* Convolution filters can introduce translucency if the sum of the
+     * weights is lower than 1.
+     */
     if (image->common.filter == PIXMAN_FILTER_CONVOLUTION)
          return FALSE;
 
-    if (image->common.repeat == PIXMAN_REPEAT_NONE)
-    {
-        if (image->common.filter != PIXMAN_FILTER_NEAREST)
-            return FALSE;
-
-        if (image->common.transform)
-            return FALSE;
-
-	/* Gradients do not necessarily cover the entire compositing area */
-	if (image->type == LINEAR || image->type == CONICAL || image->type == RADIAL)
-	    return FALSE;
-    }
-
      return TRUE;
 }
